from pathlib import Path

import pytest

import prompty

BASE_PATH = str(Path(__file__).absolute().parent.as_posix())


def test_prompty_config_local():
    p = prompty.load(f"{BASE_PATH}/prompts/sub/sub/basic.prompty")
    assert p.model.connection["type"] == "TEST_LOCAL"


@pytest.mark.asyncio
async def test_prompty_config_local_async():
    p = await prompty.load_async(f"{BASE_PATH}/prompts/sub/sub/basic.prompty")
    assert p.model.connection["type"] == "TEST_LOCAL"


def test_prompty_config_global():
    p = prompty.load(f"{BASE_PATH}/prompts/sub/basic.prompty")
    assert p.model.connection["type"] == "azure"


@pytest.mark.asyncio
async def test_prompty_config_global_async():
    p = await prompty.load_async(f"{BASE_PATH}/prompts/sub/basic.prompty")
    assert p.model.connection["type"] == "azure"


def test_prompty_config_headless():
    p = prompty.headless("embedding", ["this is the first line", "this is the second line"])
    assert p.model.connection["type"] == "FROM_CONTENT"


@pytest.mark.asyncio
async def test_prompty_config_headless_async():
    p = await prompty.headless_async("embedding", ["this is the first line", "this is the second line"])
    assert p.model.connection["type"] == "FROM_CONTENT"


# make sure the prompty path is
# relative to the current executing file
def test_prompty_relative_local():
    from tests.prompts.test import run

    p = run()
    assert p.name == "Basic Prompt"


@pytest.mark.asyncio
async def test_prompty_relative_local_async():
    from tests.prompts.test import run_async

    p = await run_async()
    assert p.name == "Basic Prompt"


def test_prompty_relative():
    from tests.prompts.sub.sub.test import run

    p = run()
    assert p.name == "Prompt with complex context"


@pytest.mark.asyncio
async def test_prompty_relative_async():
    from tests.prompts.sub.sub.test import run_async

    p = await run_async()
    assert p.name == "Prompt with complex context"



import json
import typing
from collections.abc import AsyncIterator, Iterator

from openai.types.chat.chat_completion import ChatCompletion
from openai.types.completion import Completion
from openai.types.create_embedding_response import CreateEmbeddingResponse
from openai.types.images_response import ImagesResponse

from ..core import AsyncPromptyStream, Prompty, PromptyStream, ToolCall
from ..invoker import Invoker, InvokerFactory


@InvokerFactory.register_processor("azure")
@InvokerFactory.register_processor("azure_openai")
@InvokerFactory.register_processor("azure_beta")
@InvokerFactory.register_processor("azure_openai_beta")
class AzureOpenAIProcessor(Invoker):
    """Azure OpenAI Processor"""

    def __init__(self, prompty: Prompty) -> None:
        super().__init__(prompty)

    def invoke(self, data: typing.Any) -> typing.Union[
        str,
        list[typing.Union[str, None]],
        list[ToolCall],
        list[float],
        list[list[float]],
        PromptyStream,
        None,
    ]:
        """Invoke the OpenAI/Azure API

        Parameters
        ----------
        data : any
            The data to send to the OpenAI/Azure API

        Returns
        -------
        any
            The response from the OpenAI/Azure API
        """

        if isinstance(data, ChatCompletion):
            response = data.choices[0].message
            # tool calls available in response
            if response.tool_calls:
                return [
                    ToolCall(
                        id=tool_call.id,
                        name=tool_call.function.name,
                        arguments=tool_call.function.arguments,
                    )
                    for tool_call in response.tool_calls
                ]
            else:
                if (
                    len(self.prompty.outputs) > 0
                    and response.content is not None
                    and isinstance(response.content, str)
                    and len(response.content) > 0
                ):
                    try:
                        return json.loads(response.content)
                    except json.JSONDecodeError:
                        # If the response is not JSON, return the content as is
                        return response.content
                else:
                    # add response to thread if it exists
                    thread = self.prompty.get_input("thread")
                    if thread is not None and isinstance(thread.value, list):
                        thread.value.append({"role": "assistant", "content": response.content})

                    return response.content

        elif isinstance(data, Completion):
            return data.choices[0].text
        elif isinstance(data, CreateEmbeddingResponse):
            if len(data.data) == 0:
                raise ValueError("Invalid data")
            elif len(data.data) == 1:
                return data.data[0].embedding
            else:
                return [item.embedding for item in data.data]
        elif isinstance(data, ImagesResponse):
            item: ImagesResponse = data

            if item.data is None or len(item.data) == 0:
                raise ValueError("Invalid data")
            elif len(item.data) == 1:
                return item.data[0].url if item.data[0].url else item.data[0].b64_json
            else:
                return [i.url if i.url else i.b64_json for i in item.data]

        elif isinstance(data, Iterator):

            def generator():
                for chunk in data:
                    if len(chunk.choices) == 1 and chunk.choices[0].delta.content is not None:
                        content = chunk.choices[0].delta.content
                        yield content

            return PromptyStream("AzureOpenAIProcessor", generator())
        else:
            return data

    async def invoke_async(self, data: str) -> typing.Any:
        """Invoke the Prompty Chat Parser (Async)

        Parameters
        ----------
        data : str
            The data to parse

        Returns
        -------
        str
            The parsed data
        """

        if isinstance(data, ChatCompletion):
            response = data.choices[0].message
            # tool calls available in response
            if response.tool_calls:
                return [
                    ToolCall(
                        id=tool_call.id,
                        name=tool_call.function.name,
                        arguments=tool_call.function.arguments,
                    )
                    for tool_call in response.tool_calls
                ]
            else:
                if (
                    len(self.prompty.outputs) > 0
                    and response.content is not None
                    and isinstance(response.content, str)
                    and len(response.content) > 0
                ):
                    try:
                        return json.loads(response.content)
                    except json.JSONDecodeError:
                        # If the response is not JSON, return the content as is
                        return response.content
                else:
                    return response.content

        elif isinstance(data, Completion):
            return data.choices[0].text
        elif isinstance(data, CreateEmbeddingResponse):
            if len(data.data) == 0:
                raise ValueError("Invalid data")
            elif len(data.data) == 1:
                return data.data[0].embedding
            else:
                return [item.embedding for item in data.data]
        elif isinstance(data, ImagesResponse):

            item: ImagesResponse = data

            if len(data.data) == 0:
                raise ValueError("Invalid data")
            elif len(data.data) == 1:
                return data.data[0].url if item.data[0].url else item.data[0].b64_json
            else:
                return [str(item.url) if item.url else item.b64_json for item in data.data]

        elif isinstance(data, AsyncIterator):

            async def generator():
                async for chunk in data:
                    if len(chunk.choices) == 1 and chunk.choices[0].delta.content is not None:
                        content = chunk.choices[0].delta.content
                        yield content

            return AsyncPromptyStream("AsyncAzureOpenAIProcessor", generator())
        else:
            return data



import pytest

import prompty
from prompty.common import convert_output_props


@pytest.mark.parametrize(
    "prompt",
    [
        "prompts/basic.prompty",
        "prompts/basic_json_output.prompty",
        "prompts/chat.prompty",
        "prompts/context.prompty",
        "prompts/embedding.prompty",
        "prompts/evaluation.prompty",
        "prompts/funcfile.prompty",
        "prompts/functions.prompty",
        "prompts/groundedness.prompty",
        "prompts/sub/basic.prompty",
        "prompts/sub/sub/basic.prompty",
    ],
)
def test_load(prompt: str):
    p = prompty.load(prompt)
    print(p)


@pytest.mark.parametrize(
    "prompt",
    [
        "properties/basic_array.prompty",
        "properties/basic_dictionary.prompty",
        "properties/basic_mixed.prompty",
    ],
)
def test_complex_properties(prompt: str):
    p = prompty.load(prompt)
    print(p)


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "prompt",
    [
        "properties/basic_array.prompty",
        "properties/basic_dictionary.prompty",
        "properties/basic_mixed.prompty",
    ],
)
async def test_complex_properties_async(prompt: str):
    p = await prompty.load_async(prompt)
    print(p)


@pytest.mark.parametrize(
    "prompt",
    [
        "response/structured_inline.prompty",
        "response/structured_complex.prompty",
        "response/structured_complex_other.prompty",
    ],
)
def test_complex_outputs(prompt: str):
    p = prompty.load(prompt)
    print(p)


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "prompt",
    [
        "response/structured_inline.prompty",
        "response/structured_complex.prompty",
        "response/structured_complex_other.prompty",
    ],
)
async def test_complex_outputs_async(prompt: str):
    p = await prompty.load_async(prompt)
    print(p)


@pytest.mark.parametrize(
    "prompt",
    [
        "response/structured_inline.prompty",
        "response/structured_complex.prompty",
        "response/structured_complex_other.prompty",
    ],
)
def test_convert_complex_outputs(prompt: str):
    p = prompty.load(prompt)
    o = convert_output_props(p.name, p.outputs)
    print(o)


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "prompt",
    [
        "response/structured_inline.prompty",
        "response/structured_complex.prompty",
        "response/structured_complex_other.prompty",
    ],
)
async def test_convert_complex_outputs_async(prompt: str):
    p = await prompty.load_async(prompt)
    o = convert_output_props(p.name, p.outputs)
    print(o)


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "prompt",
    [
        "prompts/basic.prompty",
        "prompts/basic_json_output.prompty",
        "prompts/chat.prompty",
        "prompts/context.prompty",
        "prompts/embedding.prompty",
        "prompts/evaluation.prompty",
        "prompts/funcfile.prompty",
        "prompts/functions.prompty",
        "prompts/groundedness.prompty",
        "prompts/sub/basic.prompty",
        "prompts/sub/sub/basic.prompty",
    ],
)
async def test_load_async(prompt: str):
    p = await prompty.load_async(prompt)
    print(p)


def test_thread_split():
    p = prompty.load("properties/thread_split.prompty")
    assert p.instructions.strip() == "before"
    assert p.additional_instructions.strip() == "after"



import json
from pathlib import Path
from unittest.mock import patch

import pytest
from dotenv import load_dotenv
from openai.types.chat.chat_completion import ChatCompletion

import prompty
from prompty.azure import AzureOpenAIProcessor
from prompty.invoker import InvokerFactory
from prompty.serverless import ServerlessProcessor
from prompty.tracer import PromptyTracer, Tracer, console_tracer, trace
from tests.fake_serverless_executor import FakeServerlessExecutor

load_dotenv()

BASE_PATH = Path(__file__).parent


def load_chat_completion(file_path: str) -> ChatCompletion:
    with open(BASE_PATH / file_path, encoding="utf-8") as f:
        return ChatCompletion.model_validate(json.load(f))


@trace
def get_current_weather(city: str, unit: str = "Celsius"):
    return f"The weather in {city} is 32 {unit}"


@trace
async def get_current_weather_async(city: str, unit: str = "Celsius"):
    return f"The weather in {city} is 32 {unit}"


@pytest.fixture(scope="module", autouse=True)
def fake_azure_executor():
    InvokerFactory.add_processor("azure", AzureOpenAIProcessor)
    InvokerFactory.add_processor("azure_openai", AzureOpenAIProcessor)
    InvokerFactory.add_executor("serverless", FakeServerlessExecutor)
    InvokerFactory.add_processor("serverless", ServerlessProcessor)

    Tracer.add("console", console_tracer)
    json_tracer = PromptyTracer()
    Tracer.add("PromptyTracer", json_tracer.tracer)


@trace
@pytest.mark.parametrize(
    "prompt",
    [
        "agent/simple_agent.prompty",
    ],
)
def test_execute_agent(prompt: str):
    p = prompty.load(prompt)
    # set the tool function
    p.set_tool_value("get_current_weather", get_current_weather)
    with patch("prompty.azure.AzureOpenAIExecutor._execute_chat_completion") as mock_create_chat:
        first = load_chat_completion(f"{prompt}.1.execution.json")
        second = load_chat_completion(f"{prompt}.2.execution.json")
        mock_create_chat.side_effect = [first, second]
        result = prompty.execute(p, inputs={"question": "What was the weather like in Tokyo?"}, merge_sample=True)
    return result


@trace
@pytest.mark.asyncio
@pytest.mark.parametrize(
    "prompt",
    [
        "agent/simple_agent.prompty",
    ],
)
async def test_execute_agent_async(prompt: str):
    p = prompty.load(prompt)
    # set the tool function
    p.set_tool_value("get_current_weather", get_current_weather_async)
    with patch("prompty.azure.AzureOpenAIExecutor._execute_chat_completion_async") as mock_create_chat:
        first = load_chat_completion(f"{prompt}.1.execution.json")
        second = load_chat_completion(f"{prompt}.2.execution.json")
        mock_create_chat.side_effect = [first, second]
        result = await prompty.execute_async(
            p, inputs={"question": "What was the weather like in Tokyo?"}, merge_sample=True
        )
    return result



import typing

from .core import OutputProperty, ToolProperty


def convert_function_tools(tools: list[ToolProperty]) -> list[dict[str, typing.Any]]:
    """Convert the tools to a list of dictionaries
    Parameters
    ----------
    tools : list[ToolProperty]
        The tools to convert
    Returns
    -------
    list[dict[str, typing.Any]]
        The converted tools
    """
    if tools:
        return [
            {
                "type": "function",
                "function": {
                    "name": tool.id,
                    "description": tool.description,
                    "parameters": {
                        "type": "object",
                        "properties": {
                            p.name: {
                                "type": p.type,
                                **({"description": p.description} if p.description else {}),
                                **({"enum": p.enum} if p.enum else {}),
                            }
                            for p in tool.parameters
                        },
                        "required": [p.name for p in tool.parameters if p.required],
                    },
                },
            }
            for tool in tools
            if tool.type == "function"
        ]
    return []


def convert_output_props(name: str, outputs: list[OutputProperty]) -> dict[str, typing.Any]:
    """Convert the tools to a list of dictionaries
    Parameters
    ----------
    tools : list[OutputProperty]
        The tools to convert
    Returns
    -------
    list[dict[str, typing.Any]]
        The converted tools
    """
    if outputs:
        return {
            "type": "json_schema",
            "json_schema": {
                "name": name,
                "strict": True,
                "schema": {
                    "type": "object",
                    "properties": {p.name: _convert_output_object(p) for p in outputs},
                    "required": [p.name for p in outputs if p.required],
                    "additionalProperties": False,
                },
            },
        }
    return {}


def _convert_output_object(output: OutputProperty) -> dict[str, typing.Any]:
    """Convert an OutputProperty to a dictionary"""
    if output.type == "array":
        if output.items is None:
            raise ValueError("Array type must have items defined")

        o = _convert_output_object(output.items)
        if "name" in o:
            o.pop("name")

        return {
            "type": "array",
            "items": o,
        }
    elif output.type == "object":
        return {
            "type": "object",
            "properties": {prop.name: _convert_output_object(prop) for prop in output.properties},
            "required": [prop.name for prop in output.properties if prop.required],
            "additionalProperties": False,
        }
    else:
        return {
            "type": output.type,
            "description": output.description,
            **({"enum": output.enum} if output.enum else {}),
        }
